
#include "base.as"
#library   lRings  "Rings.ao"

import from lRings

define BoundedPreorder:Category == Preorder with
    min: %
    max: %

define ExtremalPreorder:Category == BoundedPreorder with
    maximalNGT: % -> Generator %
    minimalNLT: % -> Generator %
    
EP ==> ExtremalPreorder
    
define Probability:Category == EP with TotalOrder with Rings with
    sqrt: % -> %
    half: % -> %
    
P:MeasureSpace with Probability == DoubleFloat add 
    Rep == DoubleFloat; import from Rep
    
    min:% == per 0.0
    max:% == per 10.0
    
    maximalNGT(x:%):Generator % == generate { yield x }
    minimalNLT(x:%):Generator % == generate { yield x }
    
    sqrt(x:%):% == error " write sqrt "
    half(x:%):% == per ( (rep x)*0.5 )
    measure(x:%):P == error "MeasureSpace error."
    
define MeasureSpace:Category == EP with 
    measure: % -> P
    
(A:MeasureSpace)*(B:MeasureSpace):MeasureSpace == add
    Rep == Record(a:A,b:B); import from Rep
    (t:TextWriter)<<(x:%):TextWriter ==  t << "(" << (rep x).a << "," << (rep x).b << ")"
    (x:%)<=(y:%):Boolean == (rep x).a <= (rep y).a and (rep x).b <= (rep y).b 
    min:% == per [ min, min ]
    max:% == per [ max, max ]
    maximalNGT(x:%):Generator % == generate
        xa:A == (rep x).a; xb:B == (rep x).b
        yield per [  xa, max ]
        yield per [ max,  xb ]
    minimalNLT(x:%):Generator % == generate
        xa:A == (rep x).a; xb:B == (rep x).b
        yield per [  xa, min ]
        yield per [ min,  xb ]
    measure(x:%):P == measure((rep x).a) * measure((rep x).b)

define MeasureSpaceHomCategory(A:MeasureSpace,B:MeasureSpace):Category == Domain with
    apply: (%,A) -> B
    doMin:  % -> A
    doMax:  % -> A
    coMin:  % -> B
    coMax:  % -> B
    minInv: % -> B -> A
    maxInv: % -> B -> A
    
    setdoMin: (%,A) -> %
    setdoMax: (%,A) -> %
    setcoMin: (%,B) -> %
    setcoMax: (%,B) -> %
    
    hom: (A->B,B->A,B->A,A,A,B,B) -> %
   
MeasureSpace(A:MeasureSpace,B:MeasureSpace):MeasureSpaceHomCategory(A,B) == add
    Rep == Record(ap:A->B,mi:B->A,ma:B->A,domin:A,domax:A,comin:B,comax:B); import from Rep
    
    apply(f:%,a:A):B == ((rep f).ap) a
    doMin(f:%):A == (rep f).domin
    doMax(f:%):A == (rep f).domax
    coMin(f:%):B == (rep f).comin
    coMax(f:%):B == (rep f).comax
    minInv(f:%)(b:B):A == ((rep f).mi) b
    maxInv(f:%)(b:B):A == ((rep f).mi) b
    
    setdoMin(f:%,a:A):% == per [ (rep f).ap, (rep f).mi, (rep f).ma,             a, (rep f).domax, (rep f).comin, (rep f).comax ]
    setdoMax(f:%,a:A):% == per [ (rep f).ap, (rep f).mi, (rep f).ma, (rep f).domin,             a, (rep f).comin, (rep f).comax ]
    setcoMin(f:%,b:B):% == per [ (rep f).ap, (rep f).mi, (rep f).ma, (rep f).domin, (rep f).domax,             b, (rep f).comax ]
    setcoMax(f:%,b:B):% == per [ (rep f).ap, (rep f).mi, (rep f).ma, (rep f).domin, (rep f).domax, (rep f).comin,             b ]

    dom(f:%):Interval A == interval ( (rep f).domin, (rep f).domax )
    
    hom(ap:A->B,mi:B->A,ma:B->A,domin:A,domax:A,comin:B,comax:B):% == per [ap,mi,ma,domin,domax,comin,comax]
      
MeasureSpaceCategory:MathCategory(MeasureSpace,MeasureSpace) == add
    id(A:MeasureSpace):MeasureSpace(A,A) == hom( (a:A):A +-> a, (a:A):A +-> a, (a:A):A +-> a, min$A, max$A, min$A, max$A  )
    compose(A:MeasureSpace,B:MeasureSpace,C:MeasureSpace)(f:MeasureSpace(A,B),g:MeasureSpace(B,C)):MeasureSpace(A,C) == 
        hom  ( (a:A):C +-> g f a, _
               (c:C):A +-> minInv(f)( (minInv)(g)(c) ), _
               (c:C):A +-> maxInv(f)( (maxInv)(g)(c) ), min$A, max$A, min$C, max$C )

Interval(A:MeasureSpace):MeasureSpace with 
    min: % -> A
    max: % -> A
    interval: (A,A) -> %
    if A has Probability then Probability
    if A has Probability then 
        **: (%,A) -> %
== add 
    Rep == Record(l:A,h:A); import from Rep;
    (t:TextWriter)<<(x:%):TextWriter == t << "(" << (rep x).l << ".." << (rep x).h << ")"
    (x:%)<=(y:%):Boolean == (rep x).l >= (rep y).l and (rep x).h <= (rep x).h
    min(x:%):A == (rep x).l
    max(x:%):A == (rep x).h
    min:% == per [ max , min ]
    max:% == per [ min , max ]
    maximalNGT(x:%):Generator % == generate
        xl:A == (rep x).l; xh:A == (rep x).h
        yield per [  xl, max ]
        yield per [ min,  xh ] 
    minimalNLT(x:%):Generator % == generate
        xl:A == (rep x).l; xh:A == (rep x).h
        yield per [  xl, min ]
        yield per [ max,  xh ]
    interval(a1:A,a2:A):% == per [a1,a2]
    measure(x:%):P == 
       amin:A == (rep x).l
       amax:A == (rep x).h
       measure amax - measure amin 
    if A has Probability then
        0:% == per [ 0, 0 ]
        1:% == per [ 1, 1 ]
        (x:%)+(y:%):% == per [ (rep x).l + (rep y).l , (rep x).h + (rep y).h ]
        (x:%)*(y:%):% == per [ (rep x).l * (rep y).l , (rep x).h + (rep y).h ]
        -(x:%):% == per [ -(rep x).l , -(rep x).h ]
        half(x:%):% == per [ half ( (rep x).l ), half ( (rep x).h ) ]
        sqrt(x:%):% == per [ sqrt ( (rep x).l ), sqrt ( (rep x).h ) ]
    if A has Probability then
        (x:%)**(a:A):% == per [ (rep x).l * a, (rep x).h * a ]


#if foo        
        
Interval(X:MeasureSpace P,P:Probability,f:EP(X,P)):EP(Interval(P)(X),Interval P) == 
    RepX == Record(l:X,h:X); import from RepX
    RepP == Record(l:P,h:P); import from RepP
    ap(ix:Interval(P) X):Interval P ==
        (x1:X,x2:X) == explode ( ix pretend RepX )
        [f x1, f x2] pretend Interval P
    mi(ip:Interval P):Interval(P) X ==
        (p1:P,p2:P) == explode ( ip pretend RepP )
        [minInv(f)(p1),maxInv(f)(p2)] pretend Interval(P) X
    ma(ip:Interval P):Interval(P) X == 
        (p1:P,p2:P) == explode ( ip pretend RepP )
        [maxInv(f)(p1),minInv(f)(p2)] pretend Interval(P) X
    minIX == [ doMin f, doMin f ] pretend Interval(P) X
    maxIX == [ doMin f, doMax f ] pretend Interval(P) X
    minIP == [ coMin f, coMin f ] pretend Interval    P
    maxIP == [ coMin f, coMax f ] pretend Interval    P
    hom(ap,mi,ma,minIX,maxIX,minIP,maxIP)
#endif
    
Interval(A:MeasureSpace,B:MeasureSpace,f:MeasureSpace(A,B)):MeasureSpace(Interval A,Interval B) == 
    RepA == Record(l:A,h:A); import from RepA
    RepB == Record(l:B,h:B); import from RepB
    ap(ia:Interval A):Interval B == 
        (a1:A,a2:A) == explode ( ia pretend RepA )
        [f a1, f a2] pretend Interval B
    mi(ib:Interval B):Interval A == 
        (b1:B,b2:B) == explode ( ib pretend RepB )
        [minInv(f)(b1),maxInv(f)(b2)] pretend Interval A
    ma(ib:Interval B):Interval A ==
        (b1:B,b2:B) == explode ( ib pretend RepB )
        [maxInv(f)(b1),minInv(f)(b2)] pretend Interval A
    minIA == [ doMin f, doMin f ] pretend Interval A
    maxIA == [ doMin f, doMax f ] pretend Interval A
    minIB == [ coMin f, coMin f ] pretend Interval B
    maxIB == [ coMin f, coMax f ] pretend Interval B
    hom(ap,mi,ma,minIA,maxIA,minIB,maxIB)


insert(A:MeasureSpace):MeasureSpace(A,Interval A) == 
    RepA == Record(l:A,h:A); import from RepA
    ap(a:A):Interval A == [ min$A, a ] pretend Interval A
    mi(ib:Interval A):A == (ib pretend RepA).h
    ma(ib:Interval A):A == (ib pretend RepA).h
    minIA == [ max$A, min$A ] pretend Interval A
    maxIA == [ min$A, max$A ] pretend Interval A
    hom(ap,mi,ma,min$A,max$A,minIA,maxIA)
    
integral(X:MeasureSpace,f:MeasureSpace(X,P)):Generator Interval P == generate 
    Intf:MeasureSpace(Interval X,Interval P) == Interval(X,P,f)
    ix:Interval X == interval ( doMin f, doMax f )
    import from X,P,MeasureSpace(X,Interval X)
    
    yield ( Intf ix ) ** ( measure doMax f - measure doMin f ) 

#if foo
choose(X:MeasureSpace,P:Probability,f:EP(X,P)):X == 
    import from P
    phalf == half ( max$P + min$P )
    maxInv(f)(phalf)

integral(X:MeasureSpace P,P:Probability,f:EP(X,P)):Generator Interval P == generate
    Intf:EP(Interval X,Interval P) == Interval(X,P,f)
    ix:Interval X == interval(doMin f,doMax f)
    import from X,P,EP(X,Interval X)
    
    yield ( Intf ix ) ** ( measure doMax f - measure doMin f )
    
    x:X == choose(X,P,f)
    import from EP(X,Interval X)
    i(x:X):Interval X == insert(X)(x)

#endif
