
#include "base.as"
#library   lRings          "Rings.ao"
#library   lPreorders  "Preorders.ao"

import from lRings,lPreorders

define Probability:Category == TotalOrder with Rings with 
    sqrt: % -> %
    half: % -> %
    
define MeasureSpace(P:Probability):Category == Preorder with
    measure: % -> P
    
define IntSpace(P:Probability):Category == MeasureSpace P with
    value:  % -> P
    minInv: P -> %
    maxInv: P -> %

    doMin: %
    doMax: %
    coMin: P
    coMax: P
    choose: () -> %
    mid: %
    default
        coMin:P == value doMin
        coMax:P == value doMax
        mid:% == maxInv half ( coMin + coMax )
    
IntSpaceCategory(P:Probability):MathCategory IntSpace P == add

Interval(P:Probability)(X:IntSpace P):IntSpace P with
    intervalmin: X -> %
    intervalmax: X -> %
== add
    Rep == Record(l:X,h:X); import from Rep
    hi(x:%):X == (rep x).h
    lo(x:%):X == (rep x).l
    bracket(x:X,y:X):% == per [x,y]
    
    (t:TextWriter)<<(x:%):TextWriter == t << "( " << (rep x).l << ".." << (rep x).h << " )"
    (x:%)<=(y:%):Boolean == (rep x).l >= (rep y).l and (rep x).h <= (rep y).h 
    value(x:%):P == value hi x
    measure(x:%):P == measure hi x - measure lo x 
    minInv(p:P):% == [ maxInv p, minInv p ]
    maxInv(p:P):% == [ minInv p, maxInv p ]
    doMin:% == [ doMax, doMin ]
    doMax:% == [ doMin, doMax ]
    coMin:P == coMin$X
    coMax:P == coMax$X

    intervalmin(x:X):% == [doMin,     x] 
    intervalmax(x:X):% == [    x, doMax]
    
    choose():% == [choose(),doMax]

Min(X:IntSpace P):IntSpace P == X add
    doMin:% == choose()

P:Probability with
    prob: % -> DoubleFloat
    prob: DoubleFloat -> %
== DoubleFloat add
    Rep == DoubleFloat; import from Rep
    
    sqrt(x:%):% == error " "
    half(x:%):% == per ( (rep x)/2.0 )
    prob(x:%):DoubleFloat == rep x
    prob(x:DoubleFloat):% == per x

integral(X:IntSpace P):Generator Record(lo:P,hi:P) == generate
    import from X
--    print << ind << ":" << " ++++++++ " << newline
--    print << ind << ":" << "  doMin, doMax: " << doMin << "  " << doMax << newline
--    print << ind << ":" << "  measure doMin, measure doMax: " << measure doMin << "  " << measure doMax << newline
--    print << ind << ":" << "  coMin, coMax: " << coMin$X << "  " << coMax$X << newline
--    print << ind << ":" << " ------ " << newline
    yield [ (coMin$X) * (measure doMax - measure doMin ) , _
            (coMax$X) * (measure doMax - measure doMin ) ]
                
    ghi:Generator Record(lo:P,hi:P) == integral (Min(X))
    glo:Generator Record(lo:P,hi:P) == integral (Min(Interval(P)(X)))
    repeat
        step! ghi; step! glo
        rlo:Record(lo:P,hi:P) == value glo
        rhi:Record(lo:P,hi:P) == value ghi
--        print << ind << ": " << "low: " << rlo.lo << " " << rlo.hi << _
--                 "    hig: " << rhi.lo << " " << rhi.hi << newline
        yield [rlo.lo + rhi.lo, rlo.hi + rhi.hi ]

